# -*- coding: utf-8 -*-
"""
Unit tests for the kernels module.

:copyright: Copyright 2014-2022 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

import math
import unittest
import warnings

import numpy as np
import quantities as pq
import scipy.integrate as spint
from numpy.testing import assert_array_almost_equal

import elephant.kernels as kernels


class kernel_TestCase(unittest.TestCase):
    def setUp(self):
        self.kernel_types = tuple(
            kern_cls for kern_cls in kernels.__dict__.values()
            if isinstance(kern_cls, type) and
            issubclass(kern_cls, kernels.Kernel) and
            kern_cls is not kernels.Kernel and
            kern_cls is not kernels.SymmetricKernel)

    def test_error_kernels(self):
        """
        Test of various error cases in the kernels module.
        """
        self.assertRaises(
            TypeError, kernels.RectangularKernel, sigma=2.0)
        self.assertRaises(
            ValueError, kernels.RectangularKernel, sigma=-0.03 * pq.s)
        self.assertRaises(
            ValueError, kernels.AlphaKernel, sigma=2.0 * pq.ms,
            invert=2)
        rec_kernel = kernels.RectangularKernel(sigma=0.3 * pq.ms)
        self.assertRaises(
            TypeError, rec_kernel, [1, 2, 3])
        self.assertRaises(
            TypeError, rec_kernel, [1, 2, 3] * pq.V)
        kernel = kernels.Kernel(sigma=0.3 * pq.ms)
        self.assertRaises(
            NotImplementedError, kernel._evaluate, [1, 2, 3] * pq.V)
        self.assertRaises(
            NotImplementedError, kernel.boundary_enclosing_area_fraction,
            fraction=0.9)
        self.assertRaises(TypeError,
                          rec_kernel.boundary_enclosing_area_fraction, [1, 2])
        self.assertRaises(ValueError,
                          rec_kernel.boundary_enclosing_area_fraction, -10)
        self.assertEqual(kernel.is_symmetric(), False)
        self.assertEqual(rec_kernel.is_symmetric(), True)

    def test_alpha_kernel_extreme(self):
        alp_kernel = kernels.AlphaKernel(sigma=0.3 * pq.ms)
        quantile = alp_kernel.boundary_enclosing_area_fraction(0.9999999)
        self.assertAlmostEqual(quantile.magnitude, 4.055922083048838)

    def test_kernels_normalization(self):
        """
        Test that each kernel normalizes to area one.
        """
        sigma = 0.1 * pq.mV
        fraction = 0.9999
        kernel_resolution = sigma / 100.0
        kernel_list = [kernel_type(sigma, invert=False) for
                       kernel_type in self.kernel_types]
        for kernel in kernel_list:
            b = kernel.boundary_enclosing_area_fraction(fraction).magnitude
            n_points = int(2 * b / kernel_resolution.magnitude)
            restric_defdomain = np.linspace(
                -b, b, num=n_points) * sigma.units
            kern = kernel(restric_defdomain)
            norm = spint.cumtrapz(y=kern.magnitude,
                                  x=restric_defdomain.magnitude)[-1]
            self.assertAlmostEqual(norm, 1, delta=0.003)

    def test_kernels_stddev(self):
        """
        Test that the standard deviation calculated from the kernel (almost)
        equals the parameter sigma with which the kernel was constructed.
        """
        sigma = 0.5 * pq.s
        fraction = 0.9999
        kernel_resolution = sigma / 50.0
        for invert in (False, True):
            kernel_list = [kernel_type(sigma, invert) for
                           kernel_type in self.kernel_types]
            for kernel in kernel_list:
                b = kernel.boundary_enclosing_area_fraction(
                    fraction).magnitude
                n_points = int(2 * b / kernel_resolution.magnitude)
                restric_defdomain = np.linspace(
                    -b, b, num=n_points) * sigma.units
                kern = kernel(restric_defdomain)
                av_integr = kern * restric_defdomain
                average = spint.cumtrapz(
                    y=av_integr.magnitude,
                    x=restric_defdomain.magnitude)[-1] * sigma.units
                var_integr = (restric_defdomain - average) ** 2 * kern
                variance = spint.cumtrapz(
                    y=var_integr.magnitude,
                    x=restric_defdomain.magnitude)[-1] * sigma.units ** 2
                stddev = np.sqrt(variance)
                self.assertAlmostEqual(stddev, sigma, delta=0.01 * sigma)

    def test_kernel_boundary_enclosing(self):
        """
        Test whether the integral of the kernel with boundary taken from
        the return value of the method boundary_enclosing_area_fraction
        is (almost) equal to the input variable `fraction` of
        boundary_enclosing_area_fraction.
        """
        sigma = 0.5 * pq.s
        kernel_resolution = sigma / 500.0
        kernel_list = [kernel_type(sigma, invert=False) for
                       kernel_type in self.kernel_types]
        for fraction in np.arange(0.15, 1.0, 0.4):
            for kernel in kernel_list:
                b = kernel.boundary_enclosing_area_fraction(fraction).magnitude
                n_points = int(2 * b / kernel_resolution.magnitude)
                restric_defdomain = np.linspace(
                    -b, b, num=n_points) * sigma.units
                kern = kernel(restric_defdomain)
                frac = spint.cumtrapz(y=kern.magnitude,
                                      x=restric_defdomain.magnitude)[-1]
                self.assertAlmostEqual(frac, fraction, delta=0.002)

    def test_kernel_output_same_size(self):
        time_array = np.linspace(0, 10, num=20) * pq.s
        for kernel_type in self.kernel_types:
            kernel = kernel_type(sigma=1 * pq.s)
            kernel_points = kernel(time_array)
            self.assertEqual(len(kernel_points), len(time_array))

    def test_element_wise_only(self):
        # Test that kernel operation is applied element-wise without any
        # recurrent magic (e.g., convolution)
        np.random.seed(19)
        t_array = np.linspace(-10, 10, num=100) * pq.s
        t_shuffled = t_array.copy()
        np.random.shuffle(t_shuffled)
        for kern_cls in self.kernel_types:
            for invert in (False, True):
                kernel = kern_cls(sigma=1 * pq.s, invert=invert)
                kernel_shuffled = np.sort(kernel(t_shuffled))
                kernel_expected = np.sort(kernel(t_array))
                assert_array_almost_equal(kernel_shuffled, kernel_expected)

    def test_kernel_pdf_range(self):
        t_array = np.linspace(-10, 10, num=1000) * pq.s
        for kern_cls in self.kernel_types:
            for invert in (False, True):
                kernel = kern_cls(sigma=1 * pq.s, invert=invert)
                kernel_array = kernel(t_array)
                in_range = (kernel_array <= 1) & (kernel_array >= 0)
                self.assertTrue(in_range.all())

    def test_boundary_enclosing_area_fraction(self):
        # test that test_boundary_enclosing_area_fraction does not depend
        # on the invert
        sigma = 1 * pq.s
        fractions_test = np.linspace(0, 1, num=10, endpoint=False)
        for kern_cls in self.kernel_types:
            kernel = kern_cls(sigma=sigma, invert=False)
            kernel_inverted = kern_cls(sigma=sigma, invert=True)
            for fraction in fractions_test:
                self.assertAlmostEqual(
                    kernel.boundary_enclosing_area_fraction(fraction),
                    kernel_inverted.boundary_enclosing_area_fraction(fraction)
                )

    def test_icdf(self):
        sigma = 1 * pq.s
        fractions_test = np.linspace(0, 1, num=10, endpoint=False)
        for kern_cls in self.kernel_types:
            kernel = kern_cls(sigma=sigma, invert=False)
            kernel_inverted = kern_cls(sigma=sigma, invert=True)
            for fraction in fractions_test:
                # ICDF(0) for several kernels produces -inf
                # of fsolve complains about stuck at local optima
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore', RuntimeWarning)
                    icdf = kernel.icdf(fraction)
                    icdf_inverted = kernel_inverted.icdf(fraction)
                if kernel.is_symmetric():
                    self.assertAlmostEqual(icdf, icdf_inverted)
                else:
                    # AlphaKernel, ExponentialKernel
                    self.assertGreaterEqual(icdf, 0 * pq.s)
                    self.assertLessEqual(icdf_inverted, 0 * pq.s)

    def test_cdf_icdf(self):
        sigma = 1 * pq.s
        fractions_test = np.linspace(0, 1, num=10, endpoint=False)
        for kern_cls in self.kernel_types:
            for invert in (False, True):
                kernel = kern_cls(sigma=sigma, invert=invert)
                for fraction in fractions_test:
                    # ICDF(0) for several kernels produces -inf
                    # of fsolve complains about stuck at local optima
                    with warnings.catch_warnings():
                        warnings.simplefilter('ignore', RuntimeWarning)
                        self.assertAlmostEqual(
                            kernel.cdf(kernel.icdf(fraction)), fraction)

    def test_icdf_cdf(self):
        sigma = 1 * pq.s
        times = np.linspace(-10, 10) * sigma.units
        for kern_cls in self.kernel_types:
            for invert in (False, True):
                kernel = kern_cls(sigma=sigma, invert=invert)
                for t in times:
                    cdf = kernel.cdf(t)
                    self.assertGreaterEqual(cdf, 0.)
                    self.assertLessEqual(cdf, 1.)
                    if 0 < cdf < 1:
                        self.assertAlmostEqual(
                            kernel.icdf(cdf), t, places=2)

    def test_icdf_at_1(self):
        sigma = 1 * pq.s
        for kern_cls in self.kernel_types:
            for invert in (False, True):
                kernel = kern_cls(sigma=sigma, invert=invert)
                if isinstance(kernel, (kernels.RectangularKernel,
                                       kernels.TriangularKernel)):
                    icdf = kernel.icdf(1.0)
                    # check finite
                    self.assertLess(np.abs(icdf.magnitude), np.inf)
                else:
                    self.assertRaises(ValueError, kernel.icdf, 1.0)

    def test_cdf_symmetric(self):
        sigma = 1 * pq.s
        cutoff = 1e2 * sigma  # a large value
        times = np.linspace(-cutoff, cutoff, num=10)
        kern_symmetric = filter(lambda kern_type: issubclass(
            kern_type, kernels.SymmetricKernel), self.kernel_types)
        for kern_cls in kern_symmetric:
            kernel = kern_cls(sigma=sigma, invert=False)
            kernel_inverted = kern_cls(sigma=sigma, invert=True)
            for t in times:
                self.assertAlmostEqual(kernel.cdf(t), kernel_inverted.cdf(t))


class KernelOldImplementation(unittest.TestCase):
    def setUp(self):
        self.kernel_types = tuple(
            kern_cls for kern_cls in kernels.__dict__.values()
            if isinstance(kern_cls, type) and
            issubclass(kern_cls, kernels.Kernel) and
            kern_cls is not kernels.Kernel and
            kern_cls is not kernels.SymmetricKernel)
        self.sigma = 1 * pq.s
        self.time_input = np.linspace(-10, 10, num=100) * self.sigma.units

    def test_triangular(self):
        def evaluate_old(t):
            t_units = t.units
            t_abs = np.abs(t.magnitude)
            tau = math.sqrt(6) * kernel.sigma.rescale(t_units).magnitude
            kernel_pdf = (t_abs < tau) * 1 / tau * (1 - t_abs / tau)
            kernel_pdf = pq.Quantity(kernel_pdf, units=1 / t_units)
            return kernel_pdf

        for invert in (False, True):
            kernel = kernels.TriangularKernel(self.sigma, invert=invert)
            assert_array_almost_equal(kernel(self.time_input),
                                      evaluate_old(self.time_input))

    def test_gaussian(self):
        def evaluate_old(t):
            t_units = t.units
            t = t.magnitude
            sigma = kernel.sigma.rescale(t_units).magnitude
            kernel_pdf = (1.0 / (math.sqrt(2.0 * math.pi) * sigma)) * np.exp(
                -0.5 * (t / sigma) ** 2)
            kernel_pdf = pq.Quantity(kernel_pdf, units=1 / t_units)
            return kernel_pdf

        for invert in (False, True):
            kernel = kernels.GaussianKernel(self.sigma, invert=invert)
            assert_array_almost_equal(kernel(self.time_input),
                                      evaluate_old(self.time_input))

    def test_laplacian(self):
        def evaluate_old(t):
            t_units = t.units
            t = t.magnitude
            tau = kernel.sigma.rescale(t_units).magnitude / math.sqrt(2)
            kernel_pdf = 1 / (2 * tau) * np.exp(-np.abs(t / tau))
            kernel_pdf = pq.Quantity(kernel_pdf, units=1 / t_units)
            return kernel_pdf

        for invert in (False, True):
            kernel = kernels.LaplacianKernel(self.sigma, invert=invert)
            assert_array_almost_equal(kernel(self.time_input),
                                      evaluate_old(self.time_input))

    def test_exponential(self):
        def evaluate_old(t):
            t_units = t.units
            t = t.magnitude
            tau = kernel.sigma.rescale(t_units).magnitude
            if not kernel.invert:
                kernel_pdf = (t >= 0) * 1 / tau * np.exp(-t / tau)
            else:
                kernel_pdf = (t <= 0) * 1 / tau * np.exp(t / tau)
            kernel_pdf = pq.Quantity(kernel_pdf, units=1 / t_units)
            return kernel_pdf

        for invert in (False, True):
            kernel = kernels.ExponentialKernel(self.sigma, invert=invert)
            assert_array_almost_equal(kernel(self.time_input),
                                      evaluate_old(self.time_input))


class KernelMedianIndex(unittest.TestCase):
    def setUp(self):
        kernel_types = tuple(
            kern_cls for kern_cls in kernels.__dict__.values()
            if isinstance(kern_cls, type) and
            issubclass(kern_cls, kernels.Kernel) and
            kern_cls is not kernels.Kernel and
            kern_cls is not kernels.SymmetricKernel)
        self.sigma = 1 * pq.s
        self.time_input = np.linspace(-10, 10, num=100) * self.sigma.units
        self.kernels = []
        for kern_cls in kernel_types:
            for invert in (False, True):
                self.kernels.append(kern_cls(self.sigma, invert=invert))

    def test_small_size(self):
        time_empty = [] * pq.s
        time_size_2 = [0, 1] * pq.s
        for kernel in self.kernels:
            self.assertRaises(ValueError, kernel.median_index, time_empty)
            median_id = kernel.median_index(time_size_2)
            self.assertEqual(median_id, 0)

    def test_not_sorted(self):
        np.random.seed(9)
        np.random.shuffle(self.time_input)
        for kernel in self.kernels:
            self.assertRaises(ValueError, kernel.median_index, self.time_input)

    def test_non_support(self):
        time_negative = np.linspace(-100, -20) * pq.s
        for kernel in self.kernels:
            if isinstance(kernel, (kernels.GaussianKernel,
                                   kernels.LaplacianKernel)):
                continue
            kernel.invert = False
            median_id = kernel.median_index(time_negative)
            self.assertEqual(median_id, len(time_negative) // 2)
            self.assertAlmostEqual(kernel.cdf(time_negative[median_id]), 0.)

    def test_old_implementation(self):
        def median_index(t):
            cumsum = kernel(t).cumsum()
            dt = (t[-1] - t[0]) / (len(t) - 1)
            quantiles = cumsum * dt
            return np.nonzero(quantiles >= 0.5)[0].min()

        for kernel in self.kernels:
            median_id = kernel.median_index(self.time_input)
            median_id_old = median_index(self.time_input)
            # the old implementation was off by 1 index, because the cumsum
            # did not start with 0 (the zero element should have been added
            # in the cumsum in old implementation).
            self.assertLessEqual(abs(median_id - median_id_old), 1)


if __name__ == '__main__':
    unittest.main()



"""
Unit tests for the Unitary Events analysis

:copyright: Copyright 2014-2022 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

import types
import unittest

import neo
import numpy as np
import quantities as pq
from numpy.testing import assert_array_equal

import elephant.unitary_event_analysis as ue
from elephant.datasets import download, ELEPHANT_TMP_DIR
from numpy.testing import assert_array_almost_equal
from elephant.spike_train_generation import StationaryPoissonProcess


class UETestCase(unittest.TestCase):

    def setUp(self):
        sts1_with_trial = [[26., 48., 78., 144., 178.],
                           [4., 45., 85., 123., 156., 185.],
                           [22., 53., 73., 88., 120., 147., 167., 193.],
                           [23., 49., 74., 116., 142., 166., 189.],
                           [5., 34., 54., 80., 108., 128., 150., 181.],
                           [18., 61., 107., 170.],
                           [62., 98., 131., 161.],
                           [37., 63., 86., 131., 168.],
                           [39., 76., 100., 127., 153., 198.],
                           [3., 35., 60., 88., 108., 141., 171., 184.],
                           [39., 170.],
                           [25., 68., 170.],
                           [19., 57., 84., 116., 157., 192.],
                           [17., 80., 131., 172.],
                           [33., 65., 124., 162., 192.],
                           [58., 87., 185.],
                           [19., 101., 174.],
                           [84., 118., 156., 198., 199.],
                           [5., 55., 67., 96., 114., 148., 172., 199.],
                           [61., 105., 131., 169., 195.],
                           [26., 96., 129., 157.],
                           [41., 85., 157., 199.],
                           [6., 30., 53., 76., 109., 142., 167., 194.],
                           [159.],
                           [6., 51., 78., 113., 154., 183.],
                           [138.],
                           [23., 59., 154., 185.],
                           [12., 14., 52., 54., 109., 145., 192.],
                           [29., 61., 84., 122., 145., 168.],
                           [26., 99.],
                           [3., 31., 55., 85., 108., 158., 191.],
                           [5., 37., 70., 119., 170.],
                           [38., 79., 117., 157., 192.],
                           [174.],
                           [114.],
                           []]
        sts2_with_trial = [[3., 119.],
                           [54., 155., 183.],
                           [35., 133.],
                           [25., 100., 176.],
                           [9., 98.],
                           [6., 97., 198.],
                           [7., 62., 148.],
                           [100., 158.],
                           [7., 62., 122., 179., 191.],
                           [125., 182.],
                           [30., 55., 127., 157., 196.],
                           [27., 70., 173.],
                           [82., 84., 198.],
                           [11., 29., 137.],
                           [5., 49., 61., 101., 142., 190.],
                           [78., 162., 178.],
                           [13., 14., 130., 172.],
                           [22.],
                           [16., 55., 109., 113., 175.],
                           [17., 33., 63., 102., 144., 189., 190.],
                           [58.],
                           [27., 30., 99., 145., 176.],
                           [10., 58., 116., 182.],
                           [14., 68., 104., 126., 162., 194.],
                           [56., 129., 196.],
                           [50., 78., 105., 152., 190., 197.],
                           [24., 66., 113., 117., 161.],
                           [9., 31., 81., 95., 136., 154.],
                           [10., 115., 185., 191.],
                           [71., 140., 157.],
                           [15., 27., 88., 102., 103., 151., 181., 188.],
                           [51., 75., 95., 134., 195.],
                           [18., 55., 75., 131., 186.],
                           [10., 16., 41., 42., 75., 127.],
                           [62., 76., 102., 145., 171., 183.],
                           [66., 71., 85., 140., 154.]]
        self.sts1_neo = [neo.SpikeTrain(
            i * pq.ms, t_stop=200 * pq.ms) for i in sts1_with_trial]
        self.sts2_neo = [neo.SpikeTrain(
            i * pq.ms, t_stop=200 * pq.ms) for i in sts2_with_trial]
        self.binary_sts = np.array([[[1, 1, 1, 1, 0],
                                     [0, 1, 1, 1, 0],
                                     [0, 1, 1, 0, 1]],
                                    [[1, 1, 1, 1, 1],
                                     [0, 1, 1, 1, 1],
                                     [1, 1, 0, 1, 0]]])

    def test_hash_default(self):
        m = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 0],
                      [1, 0, 1], [0, 1, 1], [1, 1, 1]])
        expected = np.array([77, 43, 23])
        h = ue.hash_from_pattern(m)
        self.assertTrue(np.all(expected == h))

    def test_hash_default_longpattern(self):
        m = np.zeros((100, 2))
        m[0, 0] = 1
        expected = np.array([2 ** 99, 0])
        h = ue.hash_from_pattern(m)
        self.assertTrue(np.all(expected == h))

    def test_hash_inverse_longpattern(self):
        n_patterns = 100
        m = np.random.randint(low=0, high=2, size=(n_patterns, 2))
        h = ue.hash_from_pattern(m)
        m_inv = ue.inverse_hash_from_pattern(h, N=n_patterns)
        assert_array_equal(m, m_inv)

    def test_hash_ValueError_wrong_entries(self):
        m = np.array([[0, 0, 0], [1, 0, 0], [0, 2, 0], [0, 0, 1], [1, 1, 0],
                      [1, 0, 1], [0, 1, 1], [1, 1, 1]])
        self.assertRaises(ValueError, ue.hash_from_pattern, m)

    def test_hash_base_not_two(self):
        m = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 0],
                      [1, 0, 1], [0, 1, 1], [1, 1, 1]])
        m = m.T
        base = 3
        expected = np.array([0, 9, 3, 1, 12, 10, 4, 13])
        h = ue.hash_from_pattern(m, base=base)
        self.assertTrue(np.all(expected == h))

    def test_invhash_ValueError(self):
        """
        The hash is larger than sum(2 ** range(N)).
        """
        self.assertRaises(
            ValueError, ue.inverse_hash_from_pattern, [128, 8], 4)

    def test_invhash_default_base(self):
        N = 3
        h = np.array([0, 4, 2, 1, 6, 5, 3, 7])
        expected = np.array(
            [[0, 1, 0, 0, 1, 1, 0, 1], [0, 0, 1, 0, 1, 0, 1, 1],
             [0, 0, 0, 1, 0, 1, 1, 1]])
        m = ue.inverse_hash_from_pattern(h, N)
        self.assertTrue(np.all(expected == m))

    def test_invhash_base_not_two(self):
        N = 3
        h = np.array([1, 4, 13])
        base = 3
        expected = np.array([[0, 0, 1], [0, 1, 1], [1, 1, 1]])
        m = ue.inverse_hash_from_pattern(h, N, base)
        self.assertTrue(np.all(expected == m))

    def test_invhash_shape_mat(self):
        N = 8
        h = np.array([178, 212, 232])
        expected = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1], [
            1, 1, 0], [1, 0, 1], [0, 1, 1], [1, 1, 1]])
        m = ue.inverse_hash_from_pattern(h, N)
        self.assertTrue(np.shape(m)[0] == N)

    def test_hash_invhash_consistency(self):
        m = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1],
                      [1, 1, 0], [1, 0, 1], [0, 1, 1], [1, 1, 1]])
        inv_h = ue.hash_from_pattern(m)
        m1 = ue.inverse_hash_from_pattern(inv_h, N=8)
        self.assertTrue(np.all(m == m1))

    def test_n_emp_mat_default(self):
        mat = np.array([[0, 0, 0, 1, 1], [0, 0, 0, 0, 1],
                        [1, 0, 1, 1, 1], [1, 0, 1, 1, 1]])
        pattern_hash = [3, 15]
        expected1 = np.array([2., 1.])
        expected2 = [[0, 2], [4]]
        nemp, nemp_indices = ue.n_emp_mat(mat, pattern_hash)
        self.assertTrue(np.all(nemp == expected1))
        for item_cnt, item in enumerate(nemp_indices):
            self.assertTrue(np.allclose(expected2[item_cnt], item))

    def test_n_emp_mat_sum_trial_default(self):
        mat = self.binary_sts
        pattern_hash = np.array([4, 6])
        N = 3
        expected1 = np.array([1., 3.])
        expected2 = [[[0], [3]], [[], [2, 4]]]
        n_emp, n_emp_idx = ue.n_emp_mat_sum_trial(mat, pattern_hash)
        self.assertTrue(np.all(n_emp == expected1))
        for item0_cnt, item0 in enumerate(n_emp_idx):
            for item1_cnt, item1 in enumerate(item0):
                self.assertTrue(
                    np.allclose(expected2[item0_cnt][item1_cnt], item1))

    def test_n_exp_mat_default(self):
        mat = np.array([[0, 0, 0, 1, 1], [0, 0, 0, 0, 1],
                        [1, 0, 1, 1, 1], [1, 0, 1, 1, 1]])
        pattern_hash = [3, 11]
        expected = np.array([1.536, 1.024])
        nexp = ue.n_exp_mat(mat, pattern_hash)
        self.assertTrue(np.allclose(expected, nexp))

    def test_n_exp_mat_sum_trial_default(self):
        mat = self.binary_sts
        pattern_hash = np.array([5, 6])
        expected = np.array([1.56, 2.56])
        n_exp = ue.n_exp_mat_sum_trial(mat, pattern_hash)
        self.assertTrue(np.allclose(n_exp, expected))

    def test_n_exp_mat_sum_trial_TrialAverage(self):
        mat = self.binary_sts
        pattern_hash = np.array([5, 6])
        expected = np.array([1.62, 2.52])
        n_exp = ue.n_exp_mat_sum_trial(
            mat, pattern_hash, method='analytic_TrialAverage')
        self.assertTrue(np.allclose(n_exp, expected))

    def test_n_exp_mat_sum_trial_surrogate(self):
        mat = self.binary_sts
        pattern_hash = np.array([5])
        n_exp_anal = ue.n_exp_mat_sum_trial(
            mat, pattern_hash, method='analytic_TrialAverage')
        n_exp_surr = ue.n_exp_mat_sum_trial(
            mat, pattern_hash, method='surrogate_TrialByTrial',
            n_surrogates=1000)
        self.assertLess(
            a=np.abs(n_exp_anal[0] - np.mean(n_exp_surr)) / n_exp_anal[0],
            b=0.1)

    def test_gen_pval_anal_default(self):
        mat = np.array([[[1, 1, 1, 1, 0],
                         [0, 1, 1, 1, 0],
                         [0, 1, 1, 0, 1]],

                        [[1, 1, 1, 1, 1],
                         [0, 1, 1, 1, 1],
                         [1, 1, 0, 1, 0]]])
        pattern_hash = np.array([5, 6])
        expected = np.array([1.56, 2.56])
        pval_func, n_exp = ue.gen_pval_anal(mat, pattern_hash)
        self.assertTrue(np.allclose(n_exp, expected))
        self.assertTrue(isinstance(pval_func, types.FunctionType))

    def test_jointJ_default(self):
        p_val = np.array([0.31271072, 0.01175031])
        expected = np.array([0.3419968, 1.92481736])
        self.assertTrue(np.allclose(ue.jointJ(p_val), expected))

    def test__rate_mat_avg_trial_default(self):
        mat = self.binary_sts
        expected = [0.9, 0.7, 0.6]
        self.assertTrue(np.allclose(expected, ue._rate_mat_avg_trial(mat)))

    def test__bintime(self):
        t = 13 * pq.ms
        bin_size = 3 * pq.ms
        expected = 4
        self.assertTrue(np.allclose(expected, ue._bintime(t, bin_size)))

    def test__winpos(self):
        t_start = 10 * pq.ms
        t_stop = 46 * pq.ms
        winsize = 15 * pq.ms
        winstep = 3 * pq.ms
        expected = [10., 13., 16., 19., 22., 25., 28., 31.] * pq.ms
        self.assertTrue(
            np.allclose(
                ue._winpos(t_start, t_stop, winsize,
                           winstep).rescale('ms').magnitude,
                expected.rescale('ms').magnitude))

    def test__UE_default(self):
        mat = self.binary_sts
        pattern_hash = np.array([4, 6])
        expected_S = np.array([-0.26226523, 0.04959301])
        expected_idx = [[[0], [3]], [[], [2, 4]]]
        expected_nemp = np.array([1., 3.])
        expected_nexp = np.array([1.04, 2.56])
        expected_rate = np.array([0.9, 0.7, 0.6])
        S, rate_avg, n_exp, n_emp, indices = ue._UE(mat, pattern_hash)
        self.assertTrue(np.allclose(S, expected_S))
        self.assertTrue(np.allclose(n_exp, expected_nexp))
        self.assertTrue(np.allclose(n_emp, expected_nemp))
        self.assertTrue(np.allclose(expected_rate, rate_avg))
        for item0_cnt, item0 in enumerate(indices):
            for item1_cnt, item1 in enumerate(item0):
                self.assertTrue(
                    np.allclose(expected_idx[item0_cnt][item1_cnt], item1))

    def test__UE_surrogate(self):
        mat = self.binary_sts
        pattern_hash = np.array([4])
        _, rate_avg_surr, _, n_emp_surr, indices_surr = \
            ue._UE(
                mat,
                pattern_hash,
                method='surrogate_TrialByTrial',
                n_surrogates=100)
        _, rate_avg, _, n_emp, indices = \
            ue._UE(mat, pattern_hash, method='analytic_TrialByTrial')
        self.assertTrue(np.allclose(n_emp, n_emp_surr))
        self.assertTrue(np.allclose(rate_avg, rate_avg_surr))
        for item0_cnt, item0 in enumerate(indices):
            for item1_cnt, item1 in enumerate(item0):
                self.assertTrue(
                    np.allclose(
                        indices_surr[item0_cnt][item1_cnt],
                        item1))

    def test_jointJ_window_analysis(self):
        sts1 = self.sts1_neo
        sts2 = self.sts2_neo
        data = np.vstack((sts1, sts2)).T
        win_size = 100 * pq.ms
        bin_size = 5 * pq.ms
        win_step = 20 * pq.ms
        pattern_hash = [3]
        UE_dic = ue.jointJ_window_analysis(spiketrains=data,
                                           pattern_hash=pattern_hash,
                                           bin_size=bin_size,
                                           win_size=win_size,
                                           win_step=win_step)
        expected_Js = np.array(
            [0.57953708, 0.47348757, 0.1729669,
             0.01883295, -0.21934742, -0.80608759])
        expected_n_emp = np.array(
            [9., 9., 7., 7., 6., 6.])
        expected_n_exp = np.array(
            [6.5, 6.85, 6.05, 6.6, 6.45, 8.7])
        expected_rate = np.array(
            [[0.02166667, 0.01861111],
             [0.02277778, 0.01777778],
             [0.02111111, 0.01777778],
             [0.02277778, 0.01888889],
             [0.02305556, 0.01722222],
             [0.02388889, 0.02055556]]) * pq.kHz
        expected_indecis_tril26 = [4., 4.]
        expected_indecis_tril4 = [1.]
        assert_array_almost_equal(UE_dic['Js'].squeeze(), expected_Js)
        assert_array_almost_equal(UE_dic['n_emp'].squeeze(), expected_n_emp)
        assert_array_almost_equal(UE_dic['n_exp'].squeeze(), expected_n_exp)
        assert_array_almost_equal(UE_dic['rate_avg'].squeeze(), expected_rate)
        assert_array_almost_equal(UE_dic['indices']['trial26'],
                                  expected_indecis_tril26)
        assert_array_almost_equal(UE_dic['indices']['trial4'],
                                  expected_indecis_tril4)

        # check the input parameters
        input_params = UE_dic['input_parameters']
        self.assertEqual(input_params['pattern_hash'], pattern_hash)
        self.assertEqual(input_params['bin_size'], bin_size)
        self.assertEqual(input_params['win_size'], win_size)
        self.assertEqual(input_params['win_step'], win_step)
        self.assertEqual(input_params['method'], 'analytic_TrialByTrial')
        self.assertEqual(input_params['t_start'], 0 * pq.s)
        self.assertEqual(input_params['t_stop'], 200 * pq.ms)

    @staticmethod
    def load_gdf2Neo(fname, trigger, t_pre, t_post):
        """
        load and convert the gdf file to Neo format by
        cutting and aligning around a given trigger
        # codes for trigger events (extracted from a
        # documentation of an old file after
        # contacting Dr. Alexa Rihle)
        # 700 : ST (correct) 701, 702, 703, 704*
        # 500 : ST (error =5) 501, 502, 503, 504*
        # 1000: ST (if no selec) 1001,1002,1003,1004*
        # 11  : PS 111, 112, 113, 114
        # 12  : RS 121, 122, 123, 124
        # 13  : RT 131, 132, 133, 134
        # 14  : MT 141, 142, 143, 144
        # 15  : ES 151, 152, 153, 154
        # 16  : ES 161, 162, 163, 164
        # 17  : ES 171, 172, 173, 174
        # 19  : RW 191, 192, 193, 194
        # 20  : ET 201, 202, 203, 204
        """
        data = np.loadtxt(fname)

        if trigger == 'PS_4':
            trigger_code = 114
        if trigger == 'RS_4':
            trigger_code = 124
        if trigger == 'RS':
            trigger_code = 12
        if trigger == 'ES':
            trigger_code = 15
        # specify units
        units_id = np.unique(data[:, 0][data[:, 0] < 7])
        # indecies of the trigger
        sel_tr_idx = np.where(data[:, 0] == trigger_code)[0]
        # cutting the data by aligning on the trigger
        data_tr = []
        for id_tmp in units_id:
            data_sel_units = []
            for i_cnt, i in enumerate(sel_tr_idx):
                start_tmp = data[i][1] - t_pre.magnitude
                stop_tmp = data[i][1] + t_post.magnitude
                sel_data_tmp = np.array(
                    data[np.where((data[:, 1] <= stop_tmp) &
                                  (data[:, 1] >= start_tmp))])
                sp_units_tmp = sel_data_tmp[:, 1][
                    np.where(sel_data_tmp[:, 0] == id_tmp)[0]]
                if len(sp_units_tmp) > 0:
                    aligned_time = sp_units_tmp - start_tmp
                    data_sel_units.append(neo.SpikeTrain(
                        aligned_time * pq.ms, t_start=0 * pq.ms,
                        t_stop=t_pre + t_post))
                else:
                    data_sel_units.append(neo.SpikeTrain(
                        [] * pq.ms, t_start=0 * pq.ms,
                        t_stop=t_pre + t_post))
            data_tr.append(data_sel_units)
        data_tr.reverse()
        spiketrain = np.vstack([i for i in data_tr]).T
        return spiketrain

    # Test if the result of newly implemented Unitary Events in Elephant is
    # consistent with the result of Riehle et al 1997 Science
    # (see Rostami et al (2016) [Re] Science, 3(1):1-17).
    def test_Riehle_et_al_97_UE(self):
        url = "http://raw.githubusercontent.com/ReScience-Archives/Rostami-" \
              "Ito-Denker-Gruen-2017/master/data"
        files_to_download = (
            ("extracted_data.npy", "c4903666ce8a8a31274d6b11238a5ac3"),
            ("winny131_23.gdf", "cc2958f7b4fb14dbab71e17bba49bd10")
        )
        for filename, checksum in files_to_download:
            # The files will be downloaded to ELEPHANT_TMP_DIR
            download(url=f"{url}/{filename}", checksum=checksum)

        # load spike data of figure 2 of Riehle et al 1997
        spiketrain = self.load_gdf2Neo(ELEPHANT_TMP_DIR / "winny131_23.gdf",
                                       trigger='RS_4',
                                       t_pre=1799 * pq.ms,
                                       t_post=300 * pq.ms)

        # calculating UE ...
        winsize = 100 * pq.ms
        bin_size = 5 * pq.ms
        winstep = 5 * pq.ms
        pattern_hash = [3]
        t_start = spiketrain[0][0].t_start
        t_stop = spiketrain[0][0].t_stop
        t_winpos = ue._winpos(t_start, t_stop, winsize, winstep)
        significance_level = 0.05

        UE = ue.jointJ_window_analysis(spiketrain,
                                       pattern_hash=pattern_hash,
                                       bin_size=bin_size,
                                       win_size=winsize,
                                       win_step=winstep,
                                       method='analytic_TrialAverage')
        # load extracted data from figure 2 of Riehle et al 1997
        extracted_data = np.load(ELEPHANT_TMP_DIR / 'extracted_data.npy',
                                 encoding='latin1', allow_pickle=True).item()
        Js_sig = ue.jointJ(significance_level)
        sig_idx_win = np.where(UE['Js'] >= Js_sig)[0]
        diff_UE_rep = []
        y_cnt = 0
        for trial_id in range(len(spiketrain)):
            trial_id_str = "trial{}".format(trial_id)
            indices_unique = np.unique(UE['indices'][trial_id_str])
            if len(indices_unique) > 0:
                # choose only the significant coincidences
                indices_unique_significant = []
                for j in sig_idx_win:
                    significant = indices_unique[np.where(
                        (indices_unique * bin_size >= t_winpos[j]) &
                        (indices_unique * bin_size < t_winpos[j] + winsize))]
                    indices_unique_significant.extend(significant)
                x_tmp = np.unique(indices_unique_significant) * \
                    bin_size.magnitude
                if len(x_tmp) > 0:
                    ue_trial = np.sort(extracted_data['ue'][y_cnt])
                    diff_UE_rep = np.append(
                        diff_UE_rep, x_tmp - ue_trial)
                    y_cnt += +1
        np.testing.assert_array_less(np.abs(diff_UE_rep), 0.3)

    def test_multiple_neurons(self):
        np.random.seed(12)
        spiketrains = \
            [StationaryPoissonProcess(
                rate=50 * pq.Hz, t_stop=1 * pq.s).generate_n_spiketrains(5)
             for neuron in range(3)]

        spiketrains = np.stack(spiketrains, axis=1)
        UE_dic = ue.jointJ_window_analysis(spiketrains, bin_size=5 * pq.ms,
                                           win_size=300 * pq.ms,
                                           win_step=100 * pq.ms)

        js_expected = [[0.3978179],
                       [0.08131966],
                       [-1.4239882],
                       [-0.9377029],
                       [-0.3374434],
                       [-0.2043383],
                       [-1.001536],
                       [-np.inf]]
        indices_expected = \
            {'trial3': [12, 27, 31, 34, 27, 31, 34, 136, 136, 136],
             'trial4': [4, 60, 60, 60, 117, 117, 117]}
        n_emp_expected = [[5.],
                          [4.],
                          [1.],
                          [2.],
                          [2.],
                          [2.],
                          [1.],
                          [0.]]
        n_exp_expected = [[3.5591667],
                          [3.4536111],
                          [3.3158333],
                          [3.8466666],
                          [2.370278],
                          [2.0811112],
                          [2.4011111],
                          [3.0533333]]
        rate_expected = [[[0.042, 0.03933334, 0.048]],
                         [[0.04533333, 0.038, 0.05]],
                         [[0.046, 0.04, 0.04666667]],
                         [[0.05066667, 0.042, 0.046]],
                         [[0.04466667, 0.03666667, 0.04066667]],
                         [[0.04066667, 0.03533333, 0.04333333]],
                         [[0.03933334, 0.038, 0.038]],
                         [[0.04066667, 0.04866667, 0.03666667]]] * (1. / pq.ms)
        input_parameters_expected = {'pattern_hash': [7],
                                     'bin_size': 5 * pq.ms,
                                     'win_size': 300 * pq.ms,
                                     'win_step': 100 * pq.ms,
                                     'method': 'analytic_TrialByTrial',
                                     't_start': 0 * pq.s,
                                     't_stop': 1 * pq.s, 'n_surrogates': 100}

        assert_array_almost_equal(UE_dic['Js'], js_expected)
        assert_array_almost_equal(UE_dic['n_emp'], n_emp_expected)
        assert_array_almost_equal(UE_dic['n_exp'], n_exp_expected)
        assert_array_almost_equal(UE_dic['rate_avg'], rate_expected)
        self.assertEqual(sorted(UE_dic['indices'].keys()),
                         sorted(indices_expected.keys()))
        for trial_key in indices_expected.keys():
            assert_array_equal(indices_expected[trial_key],
                               UE_dic['indices'][trial_key])
        self.assertEqual(UE_dic['input_parameters'], input_parameters_expected)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import os.path
import platform

from setuptools import setup, Extension

with open(os.path.join(os.path.dirname(__file__),
                       "elephant", "VERSION")) as version_file:
    version = version_file.read().strip()

with open("README.md") as f:
    long_description = f.read()
with open('requirements/requirements.txt') as fp:
    install_requires = fp.read().splitlines()
extras_require = {}
for extra in ['extras', 'docs', 'tests', 'tutorials', 'cuda', 'opencl']:
    with open('requirements/requirements-{0}.txt'.format(extra)) as fp:
        extras_require[extra] = fp.read()

if platform.system() == "Windows":
    fim_module = Extension(
        name='elephant.spade_src.fim',
        sources=['elephant/spade_src/src/fim.cpp'],
        include_dirs=['elephant/spade_src/include'],
        language='c++',
        libraries=[],
        extra_compile_args=[
            '-DMODULE_NAME=fim', '-DUSE_OPENMP', '-DWITH_SIG_TERM',
            '-Dfim_EXPORTS', '-fopenmp', '/std:c++17'])
elif platform.system() == "Darwin":
    fim_module = Extension(
        name='elephant.spade_src.fim',
        sources=['elephant/spade_src/src/fim.cpp'],
        include_dirs=['elephant/spade_src/include'],
        language='c++',
        libraries=['pthread', 'omp'],
        extra_compile_args=[
            '-DMODULE_NAME=fim', '-DUSE_OPENMP', '-DWITH_SIG_TERM',
            '-Dfim_EXPORTS', '-O3', '-pedantic', '-Wextra',
            '-Weffc++', '-Wunused-result', '-Werror', '-Werror=return-type',
            '-Xpreprocessor',
            '-fopenmp', '-std=gnu++17'])
elif platform.system() == "Linux":
    fim_module = Extension(
        name='elephant.spade_src.fim',
        sources=['elephant/spade_src/src/fim.cpp'],
        include_dirs=['elephant/spade_src/include'],
        language='c++',
        libraries=['pthread', 'gomp'],
        extra_compile_args=[
            '-DMODULE_NAME=fim', '-DUSE_OPENMP', '-DWITH_SIG_TERM',
            '-Dfim_EXPORTS', '-O3', '-pedantic', '-Wextra',
            '-Weffc++', '-Wunused-result', '-Werror',
            '-fopenmp', '-std=gnu++17'])

setup_kwargs = {
    "name": "elephant",
    "version": version,
    "packages": ['elephant', 'elephant.test'],
    "include_package_data": True,
    "install_requires": install_requires,
    "extras_require": extras_require,
    "author": "Elephant authors and contributors",
    "author_email": "contact@python-elephant.org",
    "description": "Elephant is a package for analysis of electrophysiology data in Python",  # noqa
    "long_description": long_description,
    "long_description_content_type": "text/markdown",
    "license": "BSD",
    "url": 'http://python-elephant.org',
    "python_requires": ">=3.7",
    "classifiers": [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering']
}
# do not compile external modules on darwin
if platform.system() in ["Windows", "Linux"]:
    setup_kwargs["ext_modules"] = [fim_module]


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""
Basic processing procedures for time series (e.g., performing a z-score of a
signal, or filtering a signal).

.. autosummary::
    :toctree: _toctree/signal_processing

    zscore
    cross_correlation_function
    butter
    wavelet_transform
    hilbert
    rauc
    derivative

:copyright: Copyright 2014-2022 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

from __future__ import division, print_function, unicode_literals

import neo
import numpy as np
import quantities as pq
import scipy.signal

from elephant.utils import deprecated_alias, check_same_units

import warnings

__all__ = [
    "zscore",
    "cross_correlation_function",
    "butter",
    "wavelet_transform",
    "hilbert",
    "rauc",
    "derivative"
]


def zscore(signal, inplace=True):
    r"""
    Apply a z-score operation to one or several `neo.AnalogSignal` objects.

    The z-score operation subtracts the mean :math:`\mu` of the signal, and
    divides by its standard deviation :math:`\sigma`:

    .. math::
         Z(x(t)) = \frac{x(t)-\mu}{\sigma}

    If a `neo.AnalogSignal` object containing multiple signals is provided,
    the z-transform is always calculated for each signal individually.

    If a list of `neo.AnalogSignal` objects is supplied, the mean and standard
    deviation are calculated across all objects of the list. Thus, all list
    elements are z-transformed by the same values of :math:`\\mu` and
    :math:`\sigma`. For a `neo.AnalogSignal` that contains multiple signals,
    each signal of the array is treated separately across list elements.
    Therefore, the number of signals must be identical for each
    `neo.AnalogSignal` object of the list.

    Parameters
    ----------
    signal : neo.AnalogSignal or list of neo.AnalogSignal
        Signals for which to calculate the z-score.
    inplace : bool, optional
        If True, the contents of the input `signal` is replaced by the
        z-transformed signal, if possible, i.e when the signal type is float.
        If the signal type is not float, an error is raised.
        If False, a copy of the original `signal` is returned.
        Default: True

    Returns
    -------
    signal_ztransformed : neo.AnalogSignal or list of neo.AnalogSignal
        The output format matches the input format: for each input
        `neo.AnalogSignal`, a corresponding `neo.AnalogSignal` is returned,
        containing the z-transformed signal with dimensionless unit.

    Raises
    ------
    ValueError
        If `inplace` is True and the type of `signal` is not float.

    Notes
    -----
    You may supply a list of `neo.AnalogSignal` objects, where each object in
    the list contains the data of one trial of the experiment, and each signal
    of the `neo.AnalogSignal` corresponds to the recordings from one specific
    electrode in a particular trial. In this scenario, you will z-transform
    the signal of each electrode separately, but transform all trials of a
    given electrode in the same way.

    Examples
    --------
    Z-transform a single `neo.AnalogSignal`, containing only a single signal.

    >>> import neo
    >>> import numpy as np
    >>> import quantities as pq
    >>> from elephant.signal_processing import zscore
    ...
    >>> a = neo.AnalogSignal(
    ...       np.array([1, 2, 3, 4, 5, 6]).reshape(-1,1) * pq.mV,
    ...       t_start=0*pq.s, sampling_rate=1000*pq.Hz)
    >>> zscore(a).as_quantity()
    [[-1.46385011]
     [-0.87831007]
     [-0.29277002]
     [ 0.29277002]
     [ 0.87831007]
     [ 1.46385011]] dimensionless

    Z-transform a single `neo.AnalogSignal` containing multiple signals.

    >>> b = neo.AnalogSignal(
    ...       np.transpose([[1, 2, 3, 4, 5, 6],
    ...                     [11, 12, 13, 14, 15, 16]]) * pq.mV,
    ...       t_start=0*pq.s, sampling_rate=1000*pq.Hz)
    >>> zscore(b).as_quantity()
    [[-1.46385011 -1.46385011]
     [-0.87831007 -0.87831007]
     [-0.29277002 -0.29277002]
     [ 0.29277002  0.29277002]
     [ 0.87831007  0.87831007]
     [ 1.46385011  1.46385011]] dimensionless

    Z-transform a list of `neo.AnalogSignal`, each one containing more than
    one signal:

    >>> c = neo.AnalogSignal(
    ...       np.transpose([[21, 22, 23, 24, 25, 26],
    ...                     [31, 32, 33, 34, 35, 36]]) * pq.mV,
    ...       t_start=0*pq.s, sampling_rate=1000*pq.Hz)
    >>> zscore([b, c])
    [<AnalogSignal(array([[-1.11669108, -1.08361877],
       [-1.0672076 , -1.04878252],
       [-1.01772411, -1.01394628],
       [-0.96824063, -0.97911003],
       [-0.91875714, -0.94427378],
       [-0.86927366, -0.90943753]]) * dimensionless, [0.0 s, 0.006 s],
       sampling rate: 1000.0 Hz)>,
       <AnalogSignal(array([[ 0.78170952,  0.84779261],
       [ 0.86621866,  0.90728682],
       [ 0.9507278 ,  0.96678104],
       [ 1.03523694,  1.02627526],
       [ 1.11974608,  1.08576948],
       [ 1.20425521,  1.1452637 ]]) * dimensionless, [0.0 s, 0.006 s],
       sampling rate: 1000.0 Hz)>]

    """
    # Transform input to a list
    if isinstance(signal, neo.AnalogSignal):
        signal = [signal]
    check_same_units(signal, object_type=neo.AnalogSignal)

    # Calculate mean and standard deviation
    signal_stacked = np.vstack(signal).magnitude
    mean = signal_stacked.mean(axis=0)
    std = signal_stacked.std(axis=0)

    signal_ztransformed = []
    for sig in signal:
        # Perform inplace operation only if array is of dtype float.
        # Otherwise, raise an error.
        if inplace and not np.issubdtype(np.float, sig.dtype):
            raise ValueError(f"Cannot perform inplace operation as the "
                             f"signal dtype is not float. Source: {sig.name}")

        sig_normalized = sig.magnitude.astype(mean.dtype, copy=not inplace)
        sig_normalized -= mean

        # items where std is zero are already zero
        np.divide(sig_normalized, std, out=sig_normalized, where=std != 0)

        if inplace:
            # Replace unit in the original array by dimensionless
            sig._dimensionality = pq.dimensionless.dimensionality
            sig_dimless = sig
        else:
            # Create new object
            sig_dimless = sig.duplicate_with_new_data(sig_normalized,
                                                      units=pq.dimensionless)
            # todo use flag once is fixed
            #      https://github.com/NeuralEnsemble/python-neo/issues/752
            sig_dimless.array_annotate(**sig.array_annotations)

        signal_ztransformed.append(sig_dimless)

    # Return single object, or list of objects
    if len(signal_ztransformed) == 1:
        signal_ztransformed = signal_ztransformed[0]
    return signal_ztransformed


@deprecated_alias(ch_pairs='channel_pairs', nlags='n_lags',
                  env='hilbert_envelope')
def cross_correlation_function(signal, channel_pairs, hilbert_envelope=False,
                               n_lags=None, scaleopt='unbiased'):
    r"""
    Computes an estimator of the cross-correlation function
    :cite:`signal-Stoica2005`.

    .. math::

             R(\tau) = \frac{1}{N-|k|} R'(\tau) \\

    where :math:`R'(\tau) = \left<x(t)y(t+\tau)\right>` in a pairwise
    manner, i.e.:

    `signal[channel_pairs[0,0]]` vs `signal[channel_pairs[0,1]]`,

    `signal[channel_pairs[1,0]]` vs `signal[channel_pairs[1,1]]`,

    and so on.

    The input time series are z-scored beforehand. `scaleopt` controls the
    choice of :math:`R_{xy}(\tau)` normalizer. Alternatively, returns the
    Hilbert envelope of :math:`R_{xy}(\tau)`, which is useful to determine the
    correlation length of oscillatory signals.

    Parameters
    ----------
    signal : (nt, nch) neo.AnalogSignal
        Signal with `nt` number of samples that contains `nch` LFP channels.
    channel_pairs : list or (n, 2) np.ndarray
        List with `n` channel pairs for which to compute cross-correlation.
        Each element of the list must contain 2 channel indices.
        If `np.ndarray`, the second axis must have dimension 2.
    hilbert_envelope : bool, optional
        If True, returns the Hilbert envelope of cross-correlation function
        result.
        Default: False
    n_lags : int, optional
        Defines the number of lags for cross-correlation function. If a `float`
        is passed, it will be rounded to the nearest integer. Number of
        samples of output is `2*n_lags+1`.
        If None, the number of samples of the output is equal to the number of
        samples of the input signal (namely `nt`).
        Default: None
    scaleopt : {'none', 'biased', 'unbiased', 'normalized', 'coeff'}, optional
        Normalization option, equivalent to matlab `xcorr(..., scaleopt)`.
        Specified as one of the following.

        * 'none': raw, unscaled cross-correlation

        .. math::
            R_{xy}(\tau)

        * 'biased': biased estimate of the cross-correlation:

        .. math::
            R_{xy,biased}(\tau) = \frac{1}{N} R_{xy}(\tau)

        * 'unbiased': unbiased estimate of the cross-correlation:

        .. math::
            R_{xy,unbiased}(\tau) = \frac{1}{N-\tau} R_{xy}(\tau)

        * 'normalized' or 'coeff': normalizes the sequence so that the
          autocorrelations at zero lag equal 1:

        .. math::
            R_{xy,coeff}(\tau) = \frac{1}{\sqrt{R_{xx}(0) R_{yy}(0)}}
                                 R_{xy}(\tau)

        Default: 'unbiased'

    Returns
    -------
    cross_corr : neo.AnalogSignal
        Shape: `[2*n_lags+1, n]`
        Pairwise cross-correlation functions for channel pairs given by
        `channel_pairs`. If `hilbert_envelope` is True, the output is the
        Hilbert envelope of the pairwise cross-correlation function. This is
        helpful to compute the correlation length for oscillating
        cross-correlation functions.

    Raises
    ------
    ValueError
        If input `signal` is not a `neo.AnalogSignal`.

        If `channel_pairs` is not a list of channel pair indices with shape
        `(n,2)`.

        If `hilbert_envelope` is not a boolean.

        If `n_lags` is not a positive integer.

        If `scaleopt` is not one of the predefined above keywords.

    Examples
    --------
    >>> import neo
    >>> import quantities as pq
    >>> import matplotlib.pyplot as plt
    >>> from elephant.signal_processing import cross_correlation_function
    >>> dt = 0.02
    >>> N = 2018
    >>> f = 0.5
    >>> t = np.arange(N)*dt
    >>> x = np.zeros((N,2))
    >>> x[:,0] = 0.2 * np.sin(2.*np.pi*f*t)
    >>> x[:,1] = 5.3 * np.cos(2.*np.pi*f*t)

    Generate neo.AnalogSignals from x and find cross-correlation

    >>> signal = neo.AnalogSignal(x, units='mV', t_start=0.*pq.ms,
    >>>     sampling_rate=1/dt*pq.Hz, dtype=float)
    >>> rho = cross_correlation_function(signal, [0,1], n_lags=150)
    >>> env = cross_correlation_function(signal, [0,1], n_lags=150,
    ...     hilbert_envelope=True)
    ...
    >>> plt.plot(rho.times, rho)
    >>> plt.plot(env.times, env) # should be equal to one
    >>> plt.show()

    """

    # Make channel_pairs a 2D array
    pairs = np.asarray(channel_pairs)
    if pairs.ndim == 1:
        pairs = np.expand_dims(pairs, axis=0)

    # Check input
    if not isinstance(signal, neo.AnalogSignal):
        raise ValueError('Input signal must be of type neo.AnalogSignal')
    if pairs.shape[1] != 2:
        raise ValueError("'channel_pairs' is not a list of channel pair "
                         "indices. Cannot define pairs for cross-correlation.")
    if not isinstance(hilbert_envelope, bool):
        raise ValueError("'hilbert_envelope' must be a boolean value")
    if n_lags is not None:
        if not isinstance(n_lags, int) or n_lags <= 0:
            raise ValueError('n_lags must be a non-negative integer')

    # z-score analog signal and store channel time series in different arrays
    # Cross-correlation will be calculated between xsig and ysig
    z_transformed = signal.magnitude - signal.magnitude.mean(axis=0)
    z_transformed = np.divide(z_transformed, signal.magnitude.std(axis=0),
                              out=z_transformed,
                              where=z_transformed != 0)
    # transpose (nch, xy, nt) -> (xy, nt, nch)
    xsig, ysig = np.transpose(z_transformed.T[pairs], (1, 2, 0))

    # Define vector of lags tau
    nt, nch = xsig.shape
    tau = np.arange(nt) - nt // 2

    # Calculate cross-correlation by taking Fourier transform of signal,
    # multiply in Fourier space, and transform back. Correct for bias due
    # to zero-padding
    xcorr = scipy.signal.fftconvolve(xsig, ysig[::-1], mode='same', axes=0)
    if scaleopt == 'biased':
        xcorr /= nt
    elif scaleopt == 'unbiased':
        normalizer = np.expand_dims(nt - np.abs(tau), axis=1)
        xcorr /= normalizer
    elif scaleopt in ('normalized', 'coeff'):
        normalizer = np.sqrt((xsig ** 2).sum(axis=0) * (ysig ** 2).sum(axis=0))
        xcorr /= normalizer
    elif scaleopt != 'none':
        raise ValueError("Invalid scaleopt mode: '{}'".format(scaleopt))

    # Calculate envelope of cross-correlation function with Hilbert transform.
    # This is useful for transient oscillatory signals.
    if hilbert_envelope:
        xcorr = np.abs(scipy.signal.hilbert(xcorr, axis=0))

    # Cut off lags outside the desired range
    if n_lags is not None:
        tau0 = np.argwhere(tau == 0).item()
        xcorr = xcorr[tau0 - n_lags: tau0 + n_lags + 1, :]

    # Return neo.AnalogSignal
    cross_corr = neo.AnalogSignal(xcorr,
                                  units='',
                                  t_start=tau[0] * signal.sampling_period,
                                  t_stop=tau[-1] * signal.sampling_period,
                                  sampling_rate=signal.sampling_rate,
                                  dtype=float)
    return cross_corr


@deprecated_alias(highpass_freq='highpass_frequency',
                  lowpass_freq='lowpass_frequency',
                  fs='sampling_frequency')
def butter(signal, highpass_frequency=None, lowpass_frequency=None, order=4,
           filter_function='filtfilt', sampling_frequency=1.0, axis=-1):
    """
    Butterworth filtering function for `neo.AnalogSignal`.

    Filter type is determined according to how values of `highpass_frequency`
    and `lowpass_frequency` are given (see "Parameters" section for details).

    Parameters
    ----------
    signal : neo.AnalogSignal or pq.Quantity or np.ndarray
        Time series data to be filtered.
        If `pq.Quantity` or `np.ndarray`, the sampling frequency should be
        given through the keyword argument `fs`.
    highpass_frequency : pq.Quantity of float, optional
        High-pass cut-off frequency. If `float`, the given value is taken as
        frequency in Hz.
        Default: None
    lowpass_frequency : pq.Quantity or float, optional
        Low-pass cut-off frequency. If `float`, the given value is taken as
        frequency in Hz.
        Filter type is determined depending on the values of
        `lowpass_frequency` and `highpass_frequency`:

        * `highpass_frequency` only (`lowpass_frequency` is None):
        highpass filter

        * `lowpass_frequency` only (`highpass_frequency` is None):
        lowpass filter

        * `highpass_frequency` < `lowpass_frequency`: bandpass filter

        * `highpass_frequency` > `lowpass_frequency`: bandstop filter

        Default: None
    order : int, optional
        Order of the Butterworth filter.
        Default: 4
    filter_function : {'filtfilt', 'lfilter', 'sosfiltfilt'}, optional
        Filtering function to be used. Available filters:

        * 'filtfilt': `scipy.signal.filtfilt`;

        * 'lfilter': `scipy.signal.lfilter`;

        * 'sosfiltfilt': `scipy.signal.sosfiltfilt`.

        In most applications 'filtfilt' should be used, because it doesn't
        bring about phase shift due to filtering. For numerically stable
        filtering, in particular higher order filters, use 'sosfiltfilt'
        (see https://github.com/NeuralEnsemble/elephant/issues/220).
        Default: 'filtfilt'
    sampling_frequency : pq.Quantity or float, optional
        The sampling frequency of the input time series. When given as
        `float`, its value is taken as frequency in Hz. When `signal` is given
        as `neo.AnalogSignal`, its attribute is used to specify the sampling
        frequency and this parameter is ignored.
        Default: 1.0
    axis : int, optional
        Axis along which filter is applied.
        Default: last axis (-1)

    Returns
    -------
    filtered_signal : neo.AnalogSignal or pq.Quantity or np.ndarray
        Filtered input data. The shape and type is identical to those of the
        input `signal`.

    Raises
    ------
    ValueError
        If `filter_function` is not one of 'lfilter', 'filtfilt',
        or 'sosfiltfilt'.

        If both `highpass_frequency` and `lowpass_frequency` are None.

    Examples
    --------
    >>> import neo
    >>> import numpy as np
    >>> import quantities as pq
    >>> from elephant.signal_processing import butter
    >>> noise = neo.AnalogSignal(np.random.normal(size=5000),
    ...     sampling_rate=1000 * pq.Hz, units='mV')
    >>> filtered_noise = butter(noise, highpass_frequency=250.0 * pq.Hz)
    >>> filtered_noise
    AnalogSignal with 1 channels of length 5000; units mV; datatype float64
    sampling rate: 1000.0 Hz
    time: 0.0 s to 5.0 s

    Let's check that the normal noise power spectrum at zero frequency is close
    to zero.

    >>> from elephant.spectral import welch_psd
    >>> freq, psd = welch_psd(filtered_noise, fs=1000.0)
    >>> psd.shape
    (1, 556)
    >>> freq[0], psd[0, 0]
    (array(0.) * Hz, array(7.21464674e-08) * mV**2/Hz)

    """
    available_filters = 'lfilter', 'filtfilt', 'sosfiltfilt'
    if filter_function not in available_filters:
        raise ValueError("Invalid `filter_function`: {filter_function}. "
                         "Available filters: {available_filters}".format(
                             filter_function=filter_function,
                             available_filters=available_filters))
    # design filter
    if hasattr(signal, 'sampling_rate'):
        sampling_frequency = signal.sampling_rate.rescale(pq.Hz).magnitude
    if isinstance(highpass_frequency, pq.quantity.Quantity):
        highpass_frequency = highpass_frequency.rescale(pq.Hz).magnitude
    if isinstance(lowpass_frequency, pq.quantity.Quantity):
        lowpass_frequency = lowpass_frequency.rescale(pq.Hz).magnitude
    Fn = sampling_frequency / 2.
    # filter type is determined according to the values of cut-off
    # frequencies
    if lowpass_frequency and highpass_frequency:
        if highpass_frequency < lowpass_frequency:
            Wn = (highpass_frequency / Fn, lowpass_frequency / Fn)
            btype = 'bandpass'
        else:
            Wn = (lowpass_frequency / Fn, highpass_frequency / Fn)
            btype = 'bandstop'
    elif lowpass_frequency:
        Wn = lowpass_frequency / Fn
        btype = 'lowpass'
    elif highpass_frequency:
        Wn = highpass_frequency / Fn
        btype = 'highpass'
    else:
        raise ValueError(
            "Either highpass_frequency or lowpass_frequency must be given"
        )
    if filter_function == 'sosfiltfilt':
        output = 'sos'
    else:
        output = 'ba'
    designed_filter = scipy.signal.butter(order, Wn, btype=btype,
                                          output=output)

    # When the input is AnalogSignal, the axis for time index (i.e. the
    # first axis) needs to be rolled to the last
    data = np.asarray(signal)
    if isinstance(signal, neo.AnalogSignal):
        data = np.rollaxis(data, 0, len(data.shape))

    # apply filter
    if filter_function == 'lfilter':
        b, a = designed_filter
        filtered_data = scipy.signal.lfilter(b=b, a=a, x=data, axis=axis)
    elif filter_function == 'filtfilt':
        b, a = designed_filter
        filtered_data = scipy.signal.filtfilt(b=b, a=a, x=data, axis=axis)
    else:
        filtered_data = scipy.signal.sosfiltfilt(sos=designed_filter,
                                                 x=data, axis=axis)

    if isinstance(signal, neo.AnalogSignal):
        filtered_data = np.rollaxis(filtered_data, -1, 0)
        signal_out = signal.duplicate_with_new_data(filtered_data)
        # todo use flag once is fixed
        #      https://github.com/NeuralEnsemble/python-neo/issues/752
        signal_out.array_annotate(**signal.array_annotations)
        return signal_out
    if isinstance(signal, pq.quantity.Quantity):
        return filtered_data * signal.units

    return filtered_data


@deprecated_alias(nco='n_cycles', freq='frequency', fs='sampling_frequency')
def wavelet_transform(signal, frequency, n_cycles=6.0, sampling_frequency=1.0,
                      zero_padding=True):
    r"""
    Compute the wavelet transform of a given signal with Morlet mother
    wavelet. The parametrization of the wavelet is based on
    :cite:`signal-Le2001_83`.

    Parameters
    ----------
    signal : (Nt, Nch) neo.AnalogSignal or np.ndarray or list
        Time series data to be wavelet-transformed. When multi-dimensional
        `np.ndarray` or list is given, the time axis must be the last
        dimension. If `neo.AnalogSignal`, `Nt` is the number of time points
        and `Nch` is the number of channels.
    frequency : float or list of float
        Center frequency of the Morlet wavelet in Hz. Multiple center
        frequencies can be given as a list, in which case the function
        computes the wavelet transforms for all the given frequencies at once.
    n_cycles : float, optional
        Size of the mother wavelet (approximate number of oscillation cycles
        within a wavelet). Corresponds to :math:`nco` in
        :cite:`signal-Le2001_83`. A larger `n_cycles` value leads to a higher
        frequency resolution and a lower temporal resolution, and vice versa.
        Typically used values are in a range of 3–8, but one should be cautious
        when using a value smaller than ~ 6, in which case the admissibility of
        the wavelet is not ensured :cite:`signal-Farge1992_395`.
        Default: 6.0
    sampling_frequency : float, optional
        Sampling rate of the input data in Hz.
        When `signal` is given as a `neo.AnalogSignal`, the sampling frequency
        is taken from its attribute and this parameter is ignored.
        Default: 1.0
    zero_padding : bool, optional
        Specifies whether the data length is extended to the least power of
        2 greater than the original length, by padding zeros to the tail, for
        speeding up the computation.
        If True, the extended part is cut out from the final result before
        returned, so that the output has the same length as the input.
        Default: True

    Returns
    -------
    signal_wt : np.ndarray
        Wavelet transform of the input data. When `frequency` was given as a
        list, the way how the wavelet transforms for different frequencies are
        returned depends on the input type:

        * when the input was a `neo.AnalogSignal`, the returned array has
          shape (`Nt`, `Nch`, `Nf`), where `Nf` = `len(freq)`, such that the
          last dimension indexes the frequencies;

        * when the input was a `np.ndarray` or list of shape
          (`a`, `b`, ..., `c`, `Nt`), the returned array has a shape
          (`a`, `b`, ..., `c`, `Nf`, `Nt`), such that the second last
          dimension indexes the frequencies.

        To summarize, `signal_wt.ndim` = `signal.ndim` + 1, with the
        additional dimension in the last axis (for `neo.AnalogSignal` input)
        or the second last axis (`np.ndarray` or list input) indexing the
        frequencies.

    Raises
    ------
    ValueError
        If `frequency` (or one of the values in `frequency` when it is a list)
        is greater than the half of `sampling_frequency`.

        If `n_cycles` is not positive.

    Notes
    -----
    `n_cycles` is related to the wavelet number :math:`w` as
    :math:`w \sim 2 \pi \frac{n_{\text{cycles}}}{6}` as defined in
    :cite:`signal-Le2001_83`.

    Examples
    --------
    >>> import neo
    >>> import numpy as np
    >>> import quantities as pq
    >>> from elephant.signal_processing import wavelet_transform
    >>> noise = neo.AnalogSignal(np.random.normal(size=7),
    ...     sampling_rate=11 * pq.Hz, units='mV')

    The wavelet frequency must be less than the half of the sampling rate;
    picking at 5 Hz.

    >>> wavelet_transform(noise, frequency=5)
    array([[-1.00890049+3.003473j  ],
       [-1.43664254-2.8389273j ],
       [ 3.02499511+0.96534578j],
       [-2.79543976+1.4581079j ],
       [ 0.94387304-2.98159518j],
       [ 1.41476471+2.77389985j],
       [-2.95996766-0.9872236j ]])

    """
    def _morlet_wavelet_ft(freq, n_cycles, fs, n):
        # Generate the Fourier transform of Morlet wavelet as defined
        # in Le van Quyen et al. J Neurosci Meth 111:83-98 (2001).
        sigma = n_cycles / (6. * freq)
        freqs = np.fft.fftfreq(n, 1.0 / fs)
        heaviside = np.array(freqs > 0., dtype=np.float)
        ft_real = np.sqrt(2 * np.pi * freq) * sigma * np.exp(
            -2 * (np.pi * sigma * (freqs - freq)) ** 2) * heaviside * fs
        ft_imag = np.zeros_like(ft_real)
        return ft_real + 1.0j * ft_imag

    data = np.asarray(signal)
    # When the input is AnalogSignal, the axis for time index (i.e. the
    # first axis) needs to be rolled to the last
    if isinstance(signal, neo.AnalogSignal):
        data = np.rollaxis(data, 0, data.ndim)

    # When the input is AnalogSignal, use its attribute to specify the
    # sampling frequency
    if hasattr(signal, 'sampling_rate'):
        sampling_frequency = signal.sampling_rate
    if isinstance(sampling_frequency, pq.quantity.Quantity):
        sampling_frequency = sampling_frequency.rescale('Hz').magnitude

    if isinstance(frequency, (list, tuple, np.ndarray)):
        freqs = np.asarray(frequency)
    else:
        freqs = np.array([frequency, ])
    if isinstance(freqs[0], pq.quantity.Quantity):
        freqs = [f.rescale('Hz').magnitude for f in freqs]

    # check whether the given central frequencies are less than the
    # Nyquist frequency of the signal
    if np.any(freqs >= sampling_frequency / 2):
        raise ValueError("'frequency' elements must be less than the half of "
                         "the 'sampling_frequency' ({}) Hz"
                         .format(sampling_frequency))

    # check if n_cycles is positive
    if n_cycles <= 0:
        raise ValueError("`n_cycles` must be positive")

    n_orig = data.shape[-1]
    if zero_padding:
        n = 2 ** (int(np.log2(n_orig)) + 1)
    else:
        n = n_orig

    # generate Morlet wavelets (in the frequency domain)
    wavelet_fts = np.empty([len(freqs), n], dtype=np.complex)
    for i, f in enumerate(freqs):
        wavelet_fts[i] = _morlet_wavelet_ft(f, n_cycles, sampling_frequency, n)

    # perform wavelet transform by convoluting the signal with the wavelets
    if data.ndim == 1:
        data = np.expand_dims(data, 0)
    data = np.expand_dims(data, data.ndim - 1)
    data = np.fft.ifft(np.fft.fft(data, n) * wavelet_fts)
    signal_wt = data[..., 0:n_orig]

    # reshape the result array according to the input
    if isinstance(signal, neo.AnalogSignal):
        signal_wt = np.rollaxis(signal_wt, -1)
        if not isinstance(frequency, (list, tuple, np.ndarray)):
            signal_wt = signal_wt[..., 0]
    else:
        if signal.ndim == 1:
            signal_wt = signal_wt[0]
        if not isinstance(frequency, (list, tuple, np.ndarray)):
            signal_wt = signal_wt[..., 0, :]

    return signal_wt


@deprecated_alias(N='padding')
def hilbert(signal, padding='nextpow'):
    """
    Apply a Hilbert transform to a `neo.AnalogSignal` object in order to
    obtain its (complex) analytic signal.

    The time series of the instantaneous angle and amplitude can be obtained
    as the angle (`np.angle` function) and absolute value (`np.abs` function)
    of the complex analytic signal, respectively.

    By default, the function will zero-pad the signal to a length
    corresponding to the next higher power of 2. This will provide higher
    computational efficiency at the expense of memory. In addition, this
    circumvents a situation where, for some specific choices of the length of
    the input, `scipy.signal.hilbert` function will not terminate.

    Parameters
    ----------
    signal : neo.AnalogSignal
        Signal(s) to transform.
    padding : int, {'none', 'nextpow'}, or None, optional
        Defines whether the signal is zero-padded.
        The `padding` argument corresponds to `N` in
        `scipy.signal.hilbert(signal, N=padding)` function.
        If 'none' or None, no padding.
        If 'nextpow', zero-pad to the next length that is a power of 2.
        If it is an `int`, directly specify the length to zero-pad to
        (indicates the number of Fourier components).
        Default: 'nextpow'

    Returns
    -------
    neo.AnalogSignal
        Contains the complex analytic signal(s) corresponding to the input
        `signal`. The unit of the returned `neo.AnalogSignal` is
        dimensionless.

    Raises
    ------
    ValueError:
        If `padding` is not an integer or neither 'nextpow' nor 'none' (None).

    Examples
    --------
    Create a sine signal at 5 Hz with increasing amplitude and calculate the
    instantaneous phases:

    >>> import neo
    >>> import numpy as np
    >>> import quantities as pq
    >>> import matplotlib.pyplot as plt
    >>> from elephant.signal_processing import hilbert
    >>> t = np.arange(0, 5000) * pq.ms
    >>> f = 5. * pq.Hz
    >>> a = neo.AnalogSignal(
    ...       np.array(
    ...           (1 + t.magnitude / t[-1].magnitude) * np.sin(
    ...               2. * np.pi * f * t.rescale(pq.s))).reshape(
    ...                   (-1,1)) * pq.mV,
    ...       t_start=0*pq.s,
    ...       sampling_rate=1000*pq.Hz)
    ...
    >>> analytic_signal = hilbert(a, padding='nextpow')
    >>> angles = np.angle(analytic_signal)
    >>> amplitudes = np.abs(analytic_signal)
    >>> print(angles)
    [[-1.57079633]
     [-1.51334228]
     [-1.46047675]
     ...,
     [-1.73112977]
     [-1.68211683]
     [-1.62879501]]
    >>> plt.plot(t, angles)

    """
    # Length of input signals
    n_org = signal.shape[0]

    # Right-pad signal to desired length using the signal itself
    if isinstance(padding, int):
        # User defined padding
        n = padding
    elif padding == 'nextpow':
        # To speed up calculation of the Hilbert transform, make sure we change
        # the signal to be of a length that is a power of two. Failure to do so
        # results in computations of certain signal lengths to not finish (or
        # finish in absurd time). This might be a bug in scipy (0.16), e.g.,
        # the following code will not terminate for this value of k:
        #
        # import numpy
        # import scipy.signal
        # k=679346
        # t = np.arange(0, k) / 1000.
        # a = (1 + t / t[-1]) * np.sin(2 * np.pi * 5 * t)
        # analytic_signal = scipy.signal.hilbert(a)
        #
        # For this reason, nextpow is the default setting for now.

        n = 2 ** (int(np.log2(n_org - 1)) + 1)
    elif padding == 'none' or padding is None:
        # No padding
        n = n_org
    else:
        raise ValueError("Invalid padding '{}'.".format(padding))

    output = signal.duplicate_with_new_data(
        scipy.signal.hilbert(signal.magnitude, N=n, axis=0)[:n_org])
    # todo use flag once is fixed
    #      https://github.com/NeuralEnsemble/python-neo/issues/752
    output.array_annotate(**signal.array_annotations)
    return output / output.units


def rauc(signal, baseline=None, bin_duration=None, t_start=None, t_stop=None):
    """
    Calculate the rectified area under the curve (RAUC) for a
    `neo.AnalogSignal`.

    The signal is optionally divided into bins with duration `bin_duration`,
    and the rectified signal (absolute value) is integrated within each bin to
    find the area under the curve. The mean or median of the signal or an
    arbitrary baseline may optionally be subtracted before rectification.

    Parameters
    ----------
    signal : neo.AnalogSignal
        The signal to integrate. If `signal` contains more than one channel,
        each is integrated separately.
    baseline : pq.Quantity or {'mean', 'median'}, optional
        A factor to subtract from the signal before rectification.
        If 'mean', the mean value of the entire `signal` is subtracted on a
        channel-by-channel basis.
        If 'median', the median value of the entire `signal` is subtracted on
        a channel-by-channel basis.
        Default: None
    bin_duration : pq.Quantity, optional
        The length of time that each integration should span.
        If None, there will be only one bin spanning the entire signal
        duration.
        If `bin_duration` does not divide evenly into the signal duration, the
        end of the signal is padded with zeros to accomodate the final,
        overextending bin.
        Default: None
    t_start : pq.Quantity, optional
        Time to start the algorithm.
        If None, starts at the beginning of `signal`.
        Default: None
    t_stop : pq.Quantity, optional
        Time to end the algorithm.
        If None, ends at the last time of `signal`.
        The signal is cropped using `signal.time_slice(t_start, t_stop)` after
        baseline removal. Useful if you want the RAUC for a short section of
        the signal but want the mean or median calculation (`baseline`='mean'
        or `baseline`='median') to use the entire signal for better baseline
        estimation.
        Default: None

    Returns
    -------
    pq.Quantity or neo.AnalogSignal
        If the number of bins is 1, the returned object is a scalar or
        vector `pq.Quantity` containing a single RAUC value for each channel.
        Otherwise, the returned object is a `neo.AnalogSignal` containing the
        RAUC(s) for each bin stored as a sample, with times corresponding to
        the center of each bin. The output signal will have the same number
        of channels as the input signal.

    Raises
    ------
    ValueError
        If `signal` is not `neo.AnalogSignal`.

        If `bin_duration` is not None or `pq.Quantity`.

        If `baseline` is not None, 'mean', 'median', or `pq.Quantity`.

    See Also
    --------
    neo.AnalogSignal.time_slice : how `t_start` and `t_stop` are used

    Examples
    --------
    >>> import neo
    >>> import numpy as np
    >>> import quantities as pq
    >>> from elephant.signal_processing import rauc
    >>> signal = neo.AnalogSignal(np.arange(10), sampling_rate=20 * pq.Hz,
    ...     units='mV')
    >>> rauc(signal)
    array(2.025) * mV/Hz

    """

    if not isinstance(signal, neo.AnalogSignal):
        raise ValueError('Input signal is not a neo.AnalogSignal!')

    if baseline is None:
        pass
    elif baseline == 'mean':
        # subtract mean from each channel
        signal = signal - signal.mean(axis=0)
    elif baseline == 'median':
        # subtract median from each channel
        signal = signal - np.median(signal.as_quantity(), axis=0)
    elif isinstance(baseline, pq.Quantity):
        # subtract arbitrary baseline
        signal = signal - baseline
    else:
        raise ValueError("baseline must be either None, 'mean', 'median', or "
                         "a Quantity. Got {}".format(baseline))

    # slice the signal after subtracting baseline
    signal = signal.time_slice(t_start, t_stop)

    if bin_duration is not None:
        # from bin duration, determine samples per bin and number of bins
        if isinstance(bin_duration, pq.Quantity):
            samples_per_bin = int(
                np.round(
                    bin_duration.rescale('s') /
                    signal.sampling_period.rescale('s')))
            n_bins = int(np.ceil(signal.shape[0] / samples_per_bin))
        else:
            raise ValueError("bin_duration must be a Quantity. Got {}".format(
                bin_duration))
    else:
        # all samples in one bin
        samples_per_bin = signal.shape[0]
        n_bins = 1

    # store the actual bin duration
    bin_duration = samples_per_bin * signal.sampling_period

    # reshape into equal size bins, padding the end with zeros if necessary
    n_channels = signal.shape[1]
    sig_binned = signal.as_quantity().copy()
    sig_binned.resize(n_bins * samples_per_bin, n_channels, refcheck=False)
    sig_binned = sig_binned.reshape(n_bins, samples_per_bin, n_channels)

    # rectify and integrate over each bin
    rauc = np.trapz(np.abs(sig_binned), dx=signal.sampling_period, axis=1)

    if n_bins == 1:
        # return a single value for each channel
        return rauc.squeeze()

    # return an AnalogSignal with times corresponding to center of each bin
    t_start = signal.t_start.rescale(bin_duration.units) + bin_duration / 2
    rauc_sig = neo.AnalogSignal(rauc, t_start=t_start,
                                sampling_period=bin_duration)
    return rauc_sig


def derivative(signal):
    """
    Calculate the derivative of a `neo.AnalogSignal`.

    Parameters
    ----------
    signal : neo.AnalogSignal
        The signal to differentiate. If `signal` contains more than one
        channel, each is differentiated separately.

    Returns
    -------
    derivative_sig : neo.AnalogSignal
        The returned object is a `neo.AnalogSignal` containing the differences
        between each successive sample value of the input signal divided by
        the sampling period. Times are centered between the successive samples
        of the input. The output signal will have the same number of channels
        as the input signal.

    Raises
    ------
    TypeError
        If `signal` is not a `neo.AnalogSignal`.

    Examples
    --------
    >>> import neo
    >>> import numpy as np
    >>> import quantities as pq
    >>> from elephant.signal_processing import derivative
    >>> signal = neo.AnalogSignal([0, 3, 4, 11, -1], sampling_rate=1 * pq.Hz,
    ...     units='mV')
    >>> print(derivative(signal))
    [[  3.]
     [  1.]
     [  7.]
     [-12.]] mV*Hz
    """

    if not isinstance(signal, neo.AnalogSignal):
        raise TypeError('Input signal is not a neo.AnalogSignal!')

    derivative_sig = neo.AnalogSignal(
        np.diff(signal.as_quantity(), axis=0) / signal.sampling_period,
        t_start=signal.t_start + signal.sampling_period / 2,
        sampling_period=signal.sampling_period)

    return derivative_sig



"""
.. autosummary::
    :toctree: _toctree/utils

    is_time_quantity
    get_common_start_stop_times
    check_neo_consistency
    check_same_units
    round_binning_errors
"""

from __future__ import division, print_function, unicode_literals

import ctypes
import warnings
from functools import wraps

import neo
from neo.core.spiketrainlist import SpikeTrainList
import numpy as np
import quantities as pq


__all__ = [
    "deprecated_alias",
    "is_binary",
    "is_time_quantity",
    "get_common_start_stop_times",
    "check_neo_consistency",
    "check_same_units",
    "round_binning_errors"
]


def is_binary(array):
    """
    Parameters
    ----------
    array : np.ndarray or list

    Returns
    -------
    bool
        Whether the input array is binary or not.

    """
    array = np.asarray(array)
    return ((array == 0) | (array == 1)).all()


def deprecated_alias(**aliases):
    """
    A deprecation decorator constructor.

    Parameters
    ----------
    **aliases
        The key-value pairs of mapping old --> new argument names of a
        function.

    Returns
    -------
    callable
        A decorator for the specific mapping of deprecated argument names.

    Examples
    --------
    In the example below, `my_function(binsize)` signature is marked as
    deprecated (but still usable) and changed to `my_function(bin_size)`.

    >>> @deprecated_alias(binsize='bin_size')
    ... def my_function(bin_size):
    ...     pass

    """
    def deco(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            _rename_kwargs(func.__name__, kwargs, aliases)
            return func(*args, **kwargs)

        return wrapper

    return deco


def _rename_kwargs(func_name, kwargs, aliases):
    for old, new in aliases.items():
        if old in kwargs:
            if new in kwargs:
                raise TypeError(f"{func_name} received both '{old}' and "
                                f"'{new}'")
            warnings.warn(f"'{old}' is deprecated; use '{new}'",
                          DeprecationWarning)
            kwargs[new] = kwargs.pop(old)


def is_time_quantity(*quantities, allow_none=False):
    """
    Parameters
    ----------
    *quantities : pq.Quantity
         A scalar or array-like to check for being a Quantity with time units.
    allow_none : bool, optional
        Allow the input to be None or not.
        Default: False

    Returns
    -------
    bool
        Whether the input is a time Quantity (True) or not (False).
        If the input is None and `allow_none` is set to True, returns True.

    """
    for quantity in quantities:
        if allow_none and quantity is None:
            continue
        if not isinstance(quantity, pq.Quantity):
            return False
        if quantity.dimensionality.simplified != pq.s.dimensionality:
            return False
    return True


def get_common_start_stop_times(neo_objects):
    """
    Extracts the common `t_start` and the `t_stop` from the input neo objects.

    If a single neo object is given, its `t_start` and `t_stop` is returned.
    Otherwise, the aligned times are returned: the maximal `t_start` and
    minimal `t_stop` across `neo_objects`.

    Parameters
    ----------
    neo_objects : neo.SpikeTrain or neo.AnalogSignal or list
        A neo object or a list of neo objects that have `t_start` and `t_stop`
        attributes.

    Returns
    -------
    t_start, t_stop : pq.Quantity
        Shared start and stop times.

    Raises
    ------
    AttributeError
        If the input neo objects do not have `t_start` and `t_stop` attributes.
    ValueError
        If there is no shared interval ``[t_start, t_stop]`` across the input
        neo objects.
    """
    if hasattr(neo_objects, 't_start') and hasattr(neo_objects, 't_stop'):
        return neo_objects.t_start, neo_objects.t_stop
    try:
        t_start = max(elem.t_start for elem in neo_objects)
        t_stop = min(elem.t_stop for elem in neo_objects)
    except AttributeError:
        raise AttributeError("Input neo objects must have 't_start' and "
                             "'t_stop' attributes")
    if t_stop < t_start:
        raise ValueError(f"t_stop ({t_stop}) is smaller than t_start "
                         f"({t_start})")
    return t_start, t_stop


def check_neo_consistency(neo_objects, object_type, t_start=None,
                          t_stop=None, tolerance=1e-8):
    """
    Checks that all input neo objects share the same units, t_start, and
    t_stop.

    Parameters
    ----------
    neo_objects : list of neo.SpikeTrain or neo.AnalogSignal
        A list of neo spike trains or analog signals.
    object_type : type
        The common type.
    t_start, t_stop : pq.Quantity or None, optional
        If None, check for exact match of t_start/t_stop across the input.
    tolerance : float, optional
        The absolute affordable tolerance for the discrepancies between
        t_start/stop magnitude values across trials.
        Default : 1e-6

    Raises
    ------
    TypeError
        If input objects are not instances of the specified `object_type`.
    ValueError
        If input object units, t_start, or t_stop do not match across trials.
    """
    if not isinstance(neo_objects, (list, tuple, SpikeTrainList)):
        neo_objects = [neo_objects]
    try:
        units = neo_objects[0].units
        start = neo_objects[0].t_start.item()
        stop = neo_objects[0].t_stop.item()
    except (IndexError, AttributeError):
        raise TypeError(f"The input must be a list of {object_type.__name__}")
    if not is_time_quantity(t_start, t_stop, allow_none=True):
        raise TypeError("'t_start' and 't_stop' must be time quantities.")
    if tolerance is None:
        tolerance = 0
    for neo_obj in neo_objects:
        if not isinstance(neo_obj, object_type):
            raise TypeError("The input must be a list of "
                            f"{object_type.__name__}. Got "
                            f"{type(neo_obj).__name__}")
        if neo_obj.units != units:
            raise ValueError("The input must have the same units.")
        if t_start is None and abs(neo_obj.t_start.item() - start) > tolerance:
            raise ValueError("The input must have the same t_start.")
        if t_stop is None and abs(neo_obj.t_stop.item() - stop) > tolerance:
            raise ValueError("The input must have the same t_stop.")


def check_same_units(quantities, object_type=pq.Quantity):
    """
    Check that all input quantities are of the same type and share common
    units. Raise an error if the check is unsuccessful.

    Parameters
    ----------
    quantities : list of pq.Quantity or pq.Quantity
        A list of quantities, neo objects or a single neo object.
    object_type : type, optional
        The common type.
        Default: pq.Quantity

    Raises
    ------
    TypeError
        If input objects are not instances of the specified `object_type`.
    ValueError
        If input objects do not share common units.
    """
    if not isinstance(quantities, (list, tuple)):
        quantities = [quantities]
    try:
        units = quantities[0].units
    except (IndexError, AttributeError):
        raise TypeError(f"The input must be a list of {object_type.__name__}")
    for quantity in quantities:
        if not isinstance(quantity, object_type):
            raise TypeError("The input must be a list of "
                            f"{object_type.__name__}. Got "
                            f"{type(quantity).__name__}")
        if quantity.units != units:
            raise ValueError("The input quantities must have the same units, "
                             "which is achieved with object.rescale('ms') "
                             "operation.")


def round_binning_errors(values, tolerance=1e-8):
    """
    Round the input `values` in-place due to the machine floating point
    precision errors.

    Parameters
    ----------
    values : np.ndarray or float
        An input array or a scalar.
    tolerance : float or None, optional
        The precision error absolute tolerance; acts as ``atol`` in
        :func:`numpy.isclose` function. If None, no rounding is performed.
        Default: 1e-8

    Returns
    -------
    values : np.ndarray or int
        Corrected integer values.

    Examples
    --------
    >>> from elephant.utils import round_binning_errors
    >>> round_binning_errors(0.999999, tolerance=None)
    0
    >>> round_binning_errors(0.999999, tolerance=1e-6)
    1
    """
    if tolerance is None or tolerance == 0:
        if isinstance(values, np.ndarray):
            return values.astype(np.int32)
        return int(values)  # a scalar

    # same as '1 - (values % 1) <= tolerance' but faster
    correction_mask = 1 - tolerance <= values % 1
    if isinstance(values, np.ndarray):
        num_corrections = correction_mask.sum()
        if num_corrections > 0:
            warnings.warn(f'Correcting {num_corrections} rounding errors by '
                          f'shifting the affected spikes into the following '
                          f'bin. You can set tolerance=None to disable this '
                          'behaviour.')
            values[correction_mask] += 0.5
        return values.astype(np.int32)

    if correction_mask:
        warnings.warn('Correcting a rounding error in the calculation '
                      'of the number of bins by incrementing the value by 1. '
                      'You can set tolerance=None to disable this '
                      'behaviour.')
        values += 0.5
    return int(values)


def get_cuda_capability_major():
    """
    Extracts CUDA capability major version of the first available Nvidia GPU
    card, if detected. Otherwise, return 0.

    Returns
    -------
    int
        CUDA capability major version.
    """
    cuda_success = 0
    for libname in ('libcuda.so', 'libcuda.dylib', 'cuda.dll'):
        try:
            cuda = ctypes.CDLL(libname)
        except OSError:
            continue
        else:
            break
    else:
        # not found
        return 0
    result = cuda.cuInit(0)
    if result != cuda_success:
        return 0
    device = ctypes.c_int()
    # parse the first GPU card only
    result = cuda.cuDeviceGet(ctypes.byref(device), 0)
    if result != cuda_success:
        return 0

    cc_major = ctypes.c_int()
    cc_minor = ctypes.c_int()
    cuda.cuDeviceComputeCapability(ctypes.byref(cc_major),
                                   ctypes.byref(cc_minor),
                                   device)
    return cc_major.value
